
(*
 * OCaml Toolkit
 * 
 * by Gamall Wednesday Ida
 * 
 * gamall-ida.com
 * gamall.ida@gmail.com
 * 
 * Copyright 2007, 2008 Gamall Wednesday Ida
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *)

(** 
  Some small functions and shortcuts which I find convenient.
  
  @author Gamall Wednesday Ida
*)

(** {6 Miscellaneous} *)

(** Functions composition operator *)
let (%) f g x = f (g x)

(** Toggle an option on and off *)
let toggle br = br := not !br

(** {6 About option and references types...} *)

exception Empty_option

(** Get the contents of an option, fail if None *)
let opt_get = function
  | None -> raise Empty_option 
  | Some x -> x

(** Get the contents of an option, use a default value if None *)
let opt_getd d = function
  | None -> d
  | Some x -> x 

(** Modified Global module from Extlib library, Copyright (C) 2003 Nicolas Cannasse 

Mutable global variable.
 
	Often in OCaml you want to have a global variable, which is mutable 
    and uninitialized when declared. You can use a ['a option ref] but
	this is not very convenient. The Global module provides functions
	to easily create and manipulate such variables.
*)
module Global = struct

  (** Raised when a global variable is accessed without first having been
  assigned a value. The parameter contains the name of the global. *)  
  exception Global_not_initialized of string
  
  (** Abstract type of a global *)
  type 'a t = ('a option ref * string)
  
  (** Returns an new named empty global. The name of the global can be any
 string. It identifies the global and makes debugging easier. *)
  let empty name = ref None,name
  
  (** Retrieve the name of a global. *)
  let name = snd
  
  (** Set the global value contents. *)
  let set (r,_) v = r := Some v
  
  (** Get the global value contents - raise Global_not_initialized if not
 defined. *)
  let get (r,name) =
          match !r with
          | None -> raise (Global_not_initialized name)
          | Some v -> v

  (** Get the global value contents - return default value if not defined *)
  let getd d (r, name) = match !r with
    | None -> d
    | Some v -> v
  
  (** Reset the global value contents to undefined. *)
  let undef (r,_) = r := None
  
  (** Return [true] if the global value has been set. *)
  let isdef (r,_) = !r <> None
  
  (** Return [None] if the global is undefined, else [Some v] where v is the
  current global value contents. *)
  let opt (r,_) = !r
end

type 'a global = 'a Global.t

(** {6 Functions about lists, strings and files } *)

(** Concatenate a list of arguments, putting spaces as separators *)
let aggreg_args al = 
  let cconc a = if a <> "" then function b -> a^" "^b else function b -> a^b in
    List.fold_left cconc "" al 

(** Is this a blank ? *)
let is_blank_char = function ' '|'\t'|'\n'|'\r' -> true | _ -> false 

(** Trim white spaces  *)
let rec trim s =
  let l = String.length s in
  if l=0 then s 
  else if is_blank_char s.[0]   then trim (String.sub s 1 (l-1))
  else if is_blank_char s.[l-1] then trim (String.sub s 0 (l-1))
  else s

(** Note: The four next functions are taken from the Str library. 
  The reason why they are not in String is beyond me...

  I have modified the first two so that they always return the string instead of 
  raising exceptions, which is what I need most of the time.
*)

(** [first_chars s n] returns the first [n] characters of [s], and [s] itself if [n] is too big *)
let first_chars s n = try String.sub s 0 n with _ -> s
      
(** [last_chars s n] returns the last [n] characters of [s], and [s] itself if [n] is too big *)
let last_chars s n = try String.sub s (String.length s - n) n with _ -> s

(** [string_before s n] returns the substring of all characters of [s]
   that precede position [n] (excluding the character at
   position [n]). *)
let string_before s n = String.sub s 0 n        

(** [string_after s n] returns the substring of all characters of [s]
   that follow position [n] (including the character at
   position [n]). *)
let string_after s n = String.sub s n (String.length s - n)

(** Get an extract of a string, up to [n+3] chars, the extra three being the "..." if the string
  is trunctated *)
let snippet s n = let res = first_chars s n in
  res ^ if String.length s <= n then "" else "..."

(** Similar in purpose to [String.sub], extracts between two positions, inclusively *)
let subb s ps pe = String.sub s ps (pe-ps)

(** Add an 's' at the end of the string if [n] > 2 *)
let pluralise s n = match abs n with 0|1 -> s | _ -> s^"s"
   
(** Tail recursive version of map *)
let map_tr f l =  let rec aux acc = function
  | [] -> List.rev acc
  | hd :: tl -> aux (f hd :: acc) tl
in aux [] l
  
(** Get rid of trailing \n *)
let rec no_newline s = let l = String.length s in match l with
  | 0 -> ""
  | _ -> if s.[l-1] = '\n' || s.[l-1] = (char_of_int 13) then no_newline (String.sub s 0 (l-1)) else s;;

(** Input a file into a list, tail-recursively, transforming each line through f (not entirely by me) *)
let lines_of_file ?(f=fun x -> x) file = let chan  = (open_in file) in
  let rec input_lines_helper res =
  let sl = 
    try
      Some (input_line chan) 
    with
      End_of_file -> None 
  in
  match sl with
       None -> List.rev res
     | Some l -> input_lines_helper ((f l) :: res) 
  in 
  input_lines_helper [];;

(** Run f for each line in the file *)
let run_on_file f path = let ch = open_in path in
  try while true do f (input_line ch)
  done with End_of_file -> ()

(** {6 Some shortcuts} *)

(** {7 Printing } *)

(** Print line to stdout *)
let pl  = print_endline 
(** Print line to stderr *)
let pe  = prerr_endline
(** Print string to stdout *)
let ps  = print_string
(** = [Printf.printf]: formated printing to stdout  *)
let pf  = Printf.printf
(** = [Printf.sprintf]: formated printing to string *)
let spf = Printf.sprintf  
(** = [Printf.eprintf]: formated printing to stderr  *)
let epf = Printf.eprintf
(** = [Printf.fprintf]: formated printing to output channel  *)
let fpf = Printf.fprintf
(** = [spf]: formated printing to stdout  *)
let va  = spf

(** {7 Lists, strings and such }*)

(** Get the tail of list *)  
let tl = List.tl
(** Get the head of a list*)
let hd = List.hd

(** = [string_of_float] *)
let sof = string_of_float
(** = [float_of_string] *)
let fos = float_of_string

(** = [string_of_int] *)
let soi = string_of_int
(** = [int_of_string] *)
let ios = int_of_string

(** = [int_of_float] *)
let iof = int_of_float
(** = [float_of_int] *)
let foi = float_of_int

(** Get a string representation of a float, but without the dot if it's an integer *)
let soff f = 
  let i = int_of_float f in
  if float_of_int i -. f = 0. then soi i else sof f

(** [ios] with debug instructions *)
let iosd debug_msg x = try int_of_string x with Failure("int_of_string") -> begin
  pf "Conversion from string to number failed on token '%s'.\nReason: \"%s\"\n" x debug_msg ; 
  failwith "Cannot continue after type conversion failure." end
  
(** [ios] with zero in case of failure *)
let iosz x = try int_of_string x with Failure("int_of_string") -> 0

(** Get length of a string *)
let strlen   = String.length
(** Get length of a list *)
let listlen  = List.length
(** Get length of an array *)
let arrlen   = Array.length
  
(** DEBUG: display codes for each char in a string *)
let break_string s = let n = strlen s in
  for i = 0 to n - 1 do pf "%d " (int_of_char s.[i]) done
   

(** {7 standard channels} *)

(** [stdin] as input channel *)
let stdinc  = Pervasives.stdin
(** [stdout]  as output channel *)
let stdoutc = Pervasives.stdout
  
(** [stdin] as a file descriptor *)
let stdinf  = Unix.stdin
(** [stdout] as a file descriptor *)
let stdoutf = Unix.stdout
  
(** {6 Unix tricks and sockets} *)

(** = [Unix.handle_unix_error] *)
let hue = Unix.handle_unix_error
  
(** Open for writing at end of file *)
let open_out_append path = 
  open_out_gen [Open_wronly; Open_creat ; Open_append; Open_text] 0o666 path


(** Get a valid port number (that is, between 0 and 65535) from a string, or fail *)    
let getport p = let res = iosd "Invalid port number" p in
  if res < 0 || res > 65535 then begin
      pf "'%d' is not a valid port number. Must be in [0, 65535]\n" res;
      failwith "Incorrect port number"; end 
  else res
     

(** DEBUG: quickly display what processus I'm in *)
let pidd s = pf "I am %s (%d)\n%!" (String.uppercase s) (Unix.getpid())
  
(** Setup a pipe with channels instead of file_descr. Optionally specify if reading is non-blocking. Blocking by default. *)
let cpipe ?(nonblock=false) () = 
  let (rdfd, wrfd) = Unix.pipe () in
  if nonblock then Unix.set_nonblock rdfd;
  let rd = Unix.in_channel_of_descr rdfd in
  let wr = Unix.out_channel_of_descr wrfd in
  (rd, wr)
    
(**
  Piped double fork, leaving no zombie process behind.
 @param nb is reading non-blocking ?
 @param f  function the son will be executing. It takes two arguments: the channels to reading from the son and writing to the son, respectively.
 @return couple of channels [(channel reading from the son, channel writing to the son)]
*)
let pdfork ?(nb=false) f = 
  let getchans() = cpipe ~nonblock:nb () in
  let srd, dwr = getchans() and drd, swr = getchans() in (* crossed pipes *)
  let close_dad ()= close_in drd; close_out dwr   in (* close the ends of the dad *)
  let close_son ()= close_in srd; close_out swr   in (* close the ends of the son *)
    match Unix.fork() with
      | 0 -> (if Unix.fork() <> 0 then () else begin close_dad(); f srd swr end); exit 0
      | n ->  ignore (Unix.waitpid [] n) ; close_son(); drd, dwr
  
(** Do an [input_line] for non-blocking channels. [None] is returned if nothing is there. *)
let peek_line ic = try Some(input_line ic) with _ -> None
  
(** Display a human-legible internet address in IP:port format*)
let string_of_sockaddr = function
  | Unix.ADDR_INET (a_ip, p) -> let ip = Unix.string_of_inet_addr a_ip in va "%s:%d" ip p
  | _ -> "???"
      
(** Build a string by repeating the same pattern n times. Useful for drawing terminal lines etc *)
let rec repeat_pattern s = function
  | n when n <= 0 -> ""
  | n -> s^(repeat_pattern s (n-1))
    
(** standard separator for terminal output: double *)
let separator_double_line  = repeat_pattern "=" 80
(** standard separator for terminal output: single *)
let separator_line         = repeat_pattern "-" 80
(** standard separator for terminal output: underscore *)
let separator_solid_line   = repeat_pattern "_" 80
  
(** {6 Command-line} *)
 
(** Array of command-line arguments *)
let av = Sys.argv
(** Number of command-line arguments *)
let ac = Array.length av
(** List of command-line arguments *)
let al = Array.to_list av    

(** Incorrect command-line arguments routine: prints all arguments, insults user, and fails.  *)
let bad_args onarg = begin 
  if onarg <> "" then epf "Failed on argument `%s'!\n" onarg;
  epf "Incorrect runtime arguments ! Run '%s --help' to see help.\n\n" av.(0);  
  let i = ref 0 in Array.iter (function x -> (epf "argument %d: <%s>\n" !i x ; i := !i+1)) av ; 
end

